-- Model: Agents - Schelling's segregation model 
--
-- Author: Tiago Carneiro & Gilberto Camara
--
-- Lesson: Agent based models in TerraME.
-- 		   Before add agents to an environment, add all cellular space the will need.
--		   Model MUST have the following structure
--				1) Define space
--				2) Define behavior (agents or societies or groups)
--				3) Define time
--				4) Define environments (adding (1) cellular spaces, then (2) agents, then (3) timers )
--				5) Run simulations 
--

--------------------------------------------------------------
-- PARAMETER 
--------------------------------------------------------------

-- satisfaction treshold
LIKEME_THRESHOLD = 5	-- try 3 (more tolerance), 4 and 5 (less tolerance)

-- size of the society of agents
SOCIETY_SIZE	= 2000
SPACE_DIMENSION = 50

-- simulationn temporal extent
FINAL_TIME = 10

--------------------------------------------------------------
-- GLOBAL 
--------------------------------------------------------------

-- agent possible state
EMPTY  = 0
BLUE   = 1
RED    = 2

-- random number generator
rand = Random()

--------------------------------------------------------------
-- MODEL 
--------------------------------------------------------------

-------------
--# SPACE #
-------------

cs = CellularSpace {
	xdim = SPACE_DIMENSION
}

cs:createNeighborhood {
    strategy = "moore",
		self = false
}

forEachCell (cs, function (cell)
	cell.color = EMPTY
end)


----------------
--# BEHAVIOR #
----------------

familyAgent = Agent {
	-- initialize the agent internal state
	init = function (self)
		if rand:number(0,1) < 0.5 then
			self.state = BLUE
		else
			self.state = RED
		end
	end,

	execute = function (self)
		if not isSatisfied(self) then
			moveFamily(self)
		end
	end
}

soc = Society {
	instance = familyAgent, 
	quantity = SOCIETY_SIZE
}

function isSatisfied (agent)

	num_equal = 0 
	cell = agent:getCell()

	forEachNeighbor (cell, function (cell, neigh)
		if neigh.color == cell.color then
			num_equal = num_equal + 1
		end
	end)

	if (num_equal >= LIKEME_THRESHOLD) then
		return true
	else
		return false
	end

end

function moveFamily (family)

	--- try to get a place to go 
	newcell = cs:sample()

	-- is the new cell empty? Yes, move; No, stay
	if newcell.color == EMPTY then
		family:getCell().color = EMPTY
		family:move (newcell)
		family:getCell().color = family.state
	end
	
end

-----------
--# TIME #
-----------

function myTimePrinter( ) 
	print(t:getTime()) io.flush()
end

t = Timer {
	Event {time = 1, period = 1, action = soc},
	Event {time = 1, period = 1, action = cs},
	Event {time = 1, period = 1, action = myTimePrinter } -- try to comment this line
}

-------------------------------------------
--# ENVIRONMENT ( SCALE or VIRTUAL WORLD) #
-------------------------------------------

-- creates the virtual world 
-- before add agents to a environment, add all cellular space the will need
env = Environment {cs, soc, t}

-- how agents will be placed inside a environment
env:createPlacement {
	strategy = "random",
	max = 1 
}

-- initialize cells with the same color of agents living on it 
forEachAgent (soc, function (agent)
		agent:getCell().color = agent.state
end)

--------------------------------------------------------------
-- SIMULATION
--------------------------------------------------------------

leg = Legend {
	grouping = "uniquevalue",
	colorBar = {
		{value = EMPTY, color = "white"},
		{value = BLUE,  color = "blue"},
		{value = RED,   color = "red"}
	}
}

map = Observer {
	type = "map",
	subject = cs,
	attributes = {"color"},
	legends = {leg}
}

-- EXECUTE THE ENVIRONMENT
env:execute(FINAL_TIME)

